#------------------------------------------------------------------
# $Copyright: Copyright (c) 2014 Symantec Corporation.
# All rights reserved.
#
# THIS SOFTWARE CONTAINS CONFIDENTIAL INFORMATION AND TRADE SECRETS OF
# SYMANTEC CORPORATION.  USE, DISCLOSURE OR REPRODUCTION IS PROHIBITED
# WITHOUT THE PRIOR EXPRESS WRITTEN PERMISSION OF SYMANTEC CORPORATION.
#
# The Licensed Software and Documentation are deemed to be commercial
# computer software as defined in FAR 12.212 and subject to restricted
# rights as defined in FAR Section 52.227-19 "Commercial Computer
# Software - Restricted Rights" and DFARS 227.7202, "Rights in
# Commercial Computer Software or Commercial Computer Software
# Documentation", as applicable, and any successor regulations. Any use,
# modification, reproduction release, performance, display or disclosure
# of the Licensed Software and Documentation by the U.S. Government
# shall be solely in accordance with the terms of this Agreement.  $
#
# $Header: /project/isv-agent/VProAgents/ConnectionBroker/ConnectionBroker.pm,v 1.4 2014/06/02 10:51:41 lpatil Exp $
#
# $sAuthor: $
#
# $Date: 2014/06/02 10:51:41 $
# $Revision: 1.4 $
#
# Provides:                            ConnectionBroker
# Author:                              rajeevk
# crAgent Build:                       1.91 2009/09/07 08:42:27
# crAgent Run Date:                    12 Nov, 2009
# Modification History:
# Known bugs/caveats:
#------------------------------------------------------------------

#------------
# Be strict..
#------------
use strict;

#----------------------
# Declare the package..
#----------------------
package ConnectionBroker;

use ag_i18n_inc;

#-------------------------------------
# Include required framework modules..
#-------------------------------------
use Sys;            # Support generalized UNIX commands
use Proc;           # For calls to the process routines
use Comms;          # For calls to socket(3C) routines
use Version;        # Version display generator
use VCSagentFW;     # Primary Agent Interface to the ACC Library
use VCSDefines;     # This contains constants that VCS uses

#--------------------------------
# Include standard PERL modules..
#--------------------------------
use Fcntl;
use Socket;
use File::Path qw ( mkpath );
use File::stat;

my ( $CLUS_HOME, $sPrefix ) = ();

#------------------------
# Setup VCS/VCSone home..
#------------------------
if ( $ENV { CLUSTER_HOME } ) { # VCS 5.x/VCSOne..
  $CLUS_HOME ||= $ENV { CLUSTER_HOME } ||= "/opt/VRTSvcsone";
}
else { # VCS 4.x..
  $CLUS_HOME ||= $ENV { VCS_HOME } ||= "/opt/VRTSvcs";
}

( $sPrefix ) = $CLUS_HOME =~ /.*(vad|vcs(|one))$/;

#-----------------------------------------
# Define module-vendor specific commands..
#-----------------------------------------
my $sCmdHares  = $HARES  ||= "$CLUS_HOME/bin/hares";
my $sCmdHatype = $HATYPE ||= "$CLUS_HOME/bin/hatype";
my $sCmdVcsDecrypt = "$CLUS_HOME/sbin/${sPrefix}decrypt";
my @liRunningProcs = (); # Global..

#-----------------------------------------------------------------------
# Add our version number to the version hash and check if versions are
# requested..
#-----------------------------------------------------------------------
my $sVer = "\$Id: ConnectionBroker.pm,v 1.4 2014/06/02 10:51:41 lpatil Exp $::ETVHeader";
Version::TellVersion ( $sVer );

#--------------------------------
# Define ConnectionBroker subroutines..
#--------------------------------

=pod

------------------------------------------------------------------------------------------------------
 Function :  bool ArgsValid ( IN String   EntryPointName,
                              IN String   CBName,
                              IN String   DMHome,
                              IN String   DMBase,
                              IN String   CBHost,
                              IN Integer  CBPort,
                              IN String   InitFile,
                              IN String   DMUser,
                              IN String   Password,
                              IN String   EnvFile,
                              IN String   MonitorProgram,
                              IN Integer  SecondLevelMonitor );

 Description : Validates the ArgList passed to the entry point.

 Parameters :

 $sEntryPointName               Entry point name
 $localCBName                   Documentum connection broker name managed by cluster server agent
 $localDMHome                   Absolute path of documentum product installation
 $localDMBase                   Absolute path of documentum connection broker binaries
 $localCBHost                   Hostname of virtual host for Documentum connection broker
 $localCBPort                   The port on which the Documentum connection broker listens
 $localInitFile                 The server initialization file for the Documentum connection broker
 $localDMUser                   Unix user owner for Documentum connection broker
 $localPassword                 Encrypted password for teh documentum user.
 $localEnvFile                  Absolute path to environment file to source
 $localMonitorProgram           Full path to external monitor program
 $localSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
------------------------------------------------------------------------------------------------------
=cut

sub ArgsValid {
  #----------------------------------------
  # Set up local vars from passed in ones..
  #----------------------------------------
  my ( $sEntryPointName,
       $localCBName,
       $localDMHome,
       $localDMBase,
       $localCBHost,
       $localCBPort,
       $localInitFile,
       $localDMUser,
       $localPassword,
       $localEnvFile,
       $localMonitorProgram,
       $localSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20001, 
  "Subroutine <%s> called with:
  EntryPointName           [%s]
  CBName                   [%s]
  DMHome                   [%s]
  DMBase                   [%s]
  CBHost                   [%s]
  CBPort                   [%s]
  InitFile                 [%s]
  DMUser                   [%s]
  Password                 [%s]
  EnvFile                  [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]",
  $sMyName, $sEntryPointName, $localCBName, $localDMHome,
  $localDMBase, $localCBHost, $localCBPort, $localInitFile,
  $localDMUser, $localPassword, $localEnvFile,
  $localMonitorProgram, $localSecondLevelMonitor );

  #-------------------------------------
  # Set return code to hopeful success..
  #-------------------------------------
  my $iReturnCode = 1;

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my ( @lsMissing, %hsArgList ) = ();
  my @lsArgs =
  qw ( EntryPointName CBName DMHome DMBase CBHost CBPort InitFile 
       DMUser Password EnvFile MonitorProgram SecondLevelMonitor );

  @hsArgList { @lsArgs } = @_; # Assign..

  #----------------------------------------
  # Reqd. attributes based on entry point..
  #----------------------------------------
  my @lsOnlineArgs  = qw ( CBName DMBase DMUser EnvFile InitFile );
  my @lsOfflineArgs = qw ( CBName DMBase DMUser EnvFile );
  my @lsMonitorArgs = qw ( CBName DMUser InitFile  );
  my @lsCleanArgs   = qw ( CBName DMUser InitFile );

  my %hlsNeeded = ( 'online'  => \@lsOnlineArgs,
                    'offline' => \@lsOfflineArgs,
                    'monitor' => \@lsMonitorArgs,
                    'clean'   => \@lsCleanArgs );

  my @lsReqdArgs = @{ $hlsNeeded { $sEntryPointName } };

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20002,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return, if failed..

  #-----------------------------
  # Remove SLM file, if exists..
  #-----------------------------
  if ( $sEntryPointName ne 'monitor' ) {
    VCSagentFW::RunSecondLevelMonitor ( 1 ); # Delete..
  }

  #-------------------------------
  # Individual validation starts..
  #-------------------------------
  
  #-----------------------------------------------
  # Validation for online offline entry point(s)..
  #-----------------------------------------------
  if ( $sEntryPointName =~ /^(online|offline)$/o ) {
    #-----------------------------
    # Validate parameter EnvFile..
    #-----------------------------
    if ( ! -e $localEnvFile ) {
      &logError ( 20003,
      "<%s> EnvFile [%s] file does not exist",
      $sMyName, $localEnvFile );
      $iReturnCode = 0; # Failure..
    }
    if ( ! -e $localDMBase ) {
      &logError ( 20004,
      "<%s> DMBase [%s] does not exist",
      $sMyName, $localDMBase );
      $iReturnCode = 0; # Failure..
    }

    #--------------------------
    # Validate parameter DMUser..
    #--------------------------
    if ( ! defined getpwnam $localDMUser ) {
       &logError ( 20005,
       "<%s> DMUser [%s] is not a valid UNIX login",
       $sMyName, $localDMUser );
       $iReturnCode = 0; # Failure..
    }
  }

  #---------------------------------------
  # Validation for online entry point(s)..
  #---------------------------------------
  if ( $sEntryPointName eq "online" ) {
    #------------------------------------
    # Validate parameter MonitorProgram..
    #------------------------------------
    if ( ! -e $localInitFile ) {
      &logError ( 20006,
      "<%s> Init File [%s] for connection broker does not exist",
      $sMyName, $localInitFile );
      $iReturnCode = 0; # Failure..
    }
    if ( $localMonitorProgram =~ /(\S+)/ && # Defined..
         ! -x $1 ) {
      &logError ( 20007,
      "<%s> MonitorProgram [%s] is not executable",
      $sMyName, $1 );
      $iReturnCode = 0; # Failure..
    }
  }

  $iReturnCode ||
  &logWarn ( 20008,
  "<%s> Some of the attributes could not be validated. Returning [%s]",
  $sMyName, $iReturnCode );

  return $iReturnCode ;
}

=pod

------------------------------------------------------------------------------------------------------ 
Function : bool Start ( IN String   CBHost,
                        IN String   CBName,
                        IN Integer  CBPort,
                        IN String   DMBase,
                        IN String   DMHome,
                        IN String   DMUser,
                        IN String   EnvFile,
                        IN String   InitFile,
                        IN String   MonitorProgram,
                        IN Integer  SecondLevelMonitor );

 Description : Start logic needed for online entry piont

 Parameters :

 $sCBHost                   Hostname of virtual host for Documentum connection broker
 $sCBName                   Documentum connection broker name managed by cluster server agent
 $iCBPort                   The port on which the Documentum connection broker listens
 $sDMBase                   Absolute path of documentum connection broker binaries
 $sDMHome                   Absolute path of documentum product installation
 $sDMUser                   Unix user owner for Documentum connection broker
 $sEnvFile                  Absolute path to environment file to source
 $sInitFile                 The server initialization file for the Documentum connection broker
 $sMonitorProgram           Full path to external monitor program
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
------------------------------------------------------------------------------------------------------

=cut

sub Start {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCBHost,
       $sCBName,
       $iCBPort,
       $sDMBase,
       $sDMHome,
       $sDMUser,
       $sEnvFile,
       $sInitFile,
       $sMonitorProgram,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20009, 
  "Subroutine <%s> called with:
  CBHost                   [%s]
  CBName                   [%s]
  CBPort                   [%s]
  DMBase                   [%s]
  DMHome                   [%s]
  DMUser                   [%s]
  EnvFile                  [%s]
  InitFile                 [%s]
  SecondLevelMonitor       [%s]",
  $sMyName, $sCBHost, $sCBName, $iCBPort, $sDMBase, $sDMHome,
  $sDMUser, $sEnvFile, $sInitFile, $iSecondLevelMonitor );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sCommand, $sCommandArgs, $iTillWhen, $iExitCode, 
       $iTimeout, @lsCmdout, @liPids ) = ();

  #-----------------------------------------
  # Whack any preexisting ConnectionBroker processes..
  #-----------------------------------------
  &logTrace ( 20010,
  "<%s> Killing ConnectionBroker processes for this instance", 
  $sMyName );

  if ( ConnectionBroker::KillConnectionBrokerProcesses ( $sDMUser,
                                           $sInitFile ) ) {
    &logTrace ( 20011,
    "<%s> Processes for this instance do not exist before start-up",
    $sMyName );
  }
  else {
    &logError ( 20012,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );
    $iReturnCode = 0; # Failure..
  }
  
  #--------------------
  # Return, if failed..
  #--------------------
  $iReturnCode || return $iReturnCode;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  $iTillWhen = $^T + ConnectionBroker::GetEngineTimeout ( 1 );
  
  #------------------------------------------
  # Define start command and args variables..
  #------------------------------------------
  $sCommand     = "$sDMBase/dm_launch_$sCBName";
  $sCommandArgs = "-init_file $sInitFile";
  ( @lsCmdout, $iExitCode, $iTimeout ) = ();

  &logTrace ( 20013,
  "<%s> Starting ConnectionBroker instance with Cmd [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------
  # Request startup..
  #------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20014,
    "<%s> Insufficient time to continue boot up. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
    ConnectionBroker::ExecLocalCmd ( $sDMUser, 
                              $sEnvFile, 
                              $sCommand, 
                              $sCommandArgs, 
                              \$iExitCode, 
                              $iTimeout, 
                              \@lsCmdout ) ) {
    &logTrace ( 20015,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }

  #-------------------
  # Return if failed..
  #-------------------
  $iReturnCode || return $iReturnCode;


  #--------------------
  # Delay entry point..
  #--------------------
  my $iSleepTime = 10; # Secs..
  
  while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
    @liPids = (); # Initialise..

    sleep $iSleepTime; # Sec..

    &logTrace ( 20016,
    "<%s> Confirming start-up with [%s] sec in hand",
    $sMyName, $iTimeout - $iSleepTime );

    #---------
    # Do FLM..
    #---------
    if ( VCSDef_MonitorExit::ONLINE ==
         ConnectionBroker::GetProcState ( \@liPids,
                                   $sDMUser,
                                   $sInitFile,
                                   1 ) ) { 
      #---------------------
      # Start-up confirmed..
      #---------------------
      &logInfo ( 20017,
      "<%s> Process check confirmed start-up of this instance of ConnectionBroker server",
      $sMyName );
 
      $iReturnCode = 1; # Success..
      last;
    }
    else {
      &logTrace ( 20018,
      "<%s> Process check could not confirm start-up. Sleeping for [%s] sec",
      $sMyName, $iSleepTime );

      $iReturnCode = 0; # Failure..
    }
  }
  if ( $iSecondLevelMonitor ) {                         # If enabled..
    while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
      
      &logTrace ( 20019,
      "<%s> Confirming start-up with [%s] sec in hand by health check",
      $sMyName, $iTimeout - $iSleepTime );
      
      #--------
      # Do SLM
      #--------
      if ( VCSDef_MonitorExit::ONLINE ==                # Online after..
           ConnectionBroker::SecondLevelMonitor ( $sCBHost,    # Running SLM..
                                           $iCBPort,
                                           $sDMHome,
                                           $sDMUser,
                                           $sEnvFile,
                                           $iTillWhen ) ) { 
        #---------------------
        # Start-up confirmed..
        #---------------------
        &logInfo ( 20020,
        "<%s> Second Level Monitor Confirmed start-up 
        of this instance of ConnectionBroker server",
        $sMyName );

        $iReturnCode = 1; # Success..
        last;
      }
      else {
        &logTrace ( 20021,
        "<%s> Could not confirm start-up. Sleeping for [%s] sec",
        $sMyName, $iSleepTime );
        
        sleep $iSleepTime; # Sec..
        $iReturnCode = 0; # Failure..
      }
    }
  }
  #------------------------------------------------------
  # MonitorProgram defined..
  # Delay the entry point for monitorprogram to succedd..
  #------------------------------------------------------
  if ( $sMonitorProgram =~ /\S+/ ) {
    my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram  # Separate
                                  =~ /(\S+)\s*(.*)/;  # Cmd and args..
    $sExtCmdArgs = '' unless defined $sExtCmdArgs;    # Prevent undefs..
    if ( $sExtCmd ) {                                 # Defined..
      while ( ( $iTimeout = $iTillWhen - time ) > 10 ) {
        if ( VCSDef_MonitorExit::ONLINE ==            # and found online..
             ConnectionBroker::ExternalMonitor ( $sDMUser,
                                          $sEnvFile,
                                          $sExtCmd,
                                          $sExtCmdArgs,
                                          $iTillWhen ) ) {
          &logTrace ( 20022,
          "<%s> ConnectionBroker instance found online  after 
          executing external monitor [%s] with args [%s]",
          $sMyName, $sExtCmd, $sExtCmdArgs );
          
          $iReturnCode = 1; # Success..
          last;
        }
        else {
          &logTrace ( 20023,
          "<%s> ConnectionBroker instance NOT online after executing external monitor
          [%s] with args [%s]. Sleeping for [%s] sec",
          $sMyName, $sExtCmd, $sExtCmdArgs, $iSleepTime );
          
          sleep $iSleepTime; # Sec..
          $iReturnCode = 0;  # Failure..
        }
      }
    }
  }

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------ 
Function : bool Stop ( IN String  CBName,
                       IN String  DMBase,
                       IN String  DMUser,
                       IN String  EnvFile,
                       IN String  InitFile,
                       IN String  Password );

 Description : Stop logic needed for offline entry point

 Parameters :

 $sCBName                   Documentum connection broker name managed by cluster server agent
 $sDMBase                   Absolute path of documentum connection broker binaries
 $sDMUser                   Unix user owner for Documentum connection broker
 $sEnvFile                  Absolute path to environment file to source
 $sInitFile                 The server initialization file for the Documentum connection broker
 $sPassword                 Encrypted password for teh documentum user.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
------------------------------------------------------------------------------------------------------

=cut

sub Stop {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCBName,
       $sDMBase,
       $sDMUser,
       $sEnvFile,
       $sInitFile,
       $sPassword ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20024, 
  "Subroutine <%s> called with:
  CBName                   [%s]
  DMBase                   [%s]
  DMUser                   [%s]
  EnvFile                  [%s]
  InitFile                 [%s]
  Password                 [%s]",
  $sMyName, $sCBName, $sDMBase, $sDMUser,
  $sEnvFile, $sInitFile, $sPassword );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + ConnectionBroker::GetEngineTimeout ( 1 ); # With margin..

  #------------------------------------------
  # Declare stop command and args variables..
  #------------------------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode, 
       $iSleepTime, @lsCmdout, @liPids ) = ();
  
  #------------------------------------
  # Define stop command and arguments..
  #------------------------------------
  $sCommand = "$sDMBase/dm_stop_$sCBName";
  $sCommandArgs = "-P$sPassword";
  ( $iExitCode, $iTimeout, @lsCmdout ) = (); # Truncate.. 

  &logTrace ( 20025,
  "<%s> Stopping ConnectionBroker with Command [%s] Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  #------------------------
  # Request shutdown..
  #------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20026,
    "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = 0; # Failure..
  }
  elsif ( $iReturnCode =
          ConnectionBroker::ExecLocalCmd ( $sDMUser, 
                                    $sEnvFile, 
                                    $sCommand, 
                                    $sCommandArgs, 
                                    \$iExitCode, 
                                    $iTimeout, 
                                    \@lsCmdout,
                                    $sPassword ) ) {
    &logTrace ( 20027,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
  }


  #--------------------
  # Delay entry point..
  #--------------------
  if ( $iReturnCode ) { # Success..
    #-----------------
    # Set sleep time..
    #-----------------
    $iSleepTime = 10; # Sec..
  
    &logInfo ( 20028,
    "<%s> Delaying exit to confirm shutdown..", $sMyName );

    #--------------------
    # Delay entry point..
    #--------------------
    while ( ( $iTimeout = $iTillWhen - time ) > 20 ) {
      #-----------------------
      # Wait before retrying..
      #-----------------------
      sleep $iSleepTime; # Secs..
  
      ConnectionBroker::GetProcState ( \@liPids,
                                $sDMUser,
                                $sInitFile,
                                0 );  # Request regular..
  
      #------------------------------
      # Check for pending processes..
      #------------------------------
      if ( scalar @liPids ) {
        &logTrace ( 20029,
        "<%s> Could not confirm shutdown. Retrying in [%s] sec.
        Delaying kill with [%s] sec in hand",
        $sMyName, $iSleepTime, $iTimeout - $iSleepTime );
      }
      else {
        &logInfo ( 20030,
        "<%s> Confirmed shutdown of ConnectionBroker instance",
        $sMyName );
        last;  # Get out..
      }
    }
  }

  #-----------------------------------------
  # Whack any persisting ConnectionBroker processes..
  #-----------------------------------------
  &logTrace ( 20031,
  "<%s> Killing ConnectionBroker processes for this instance", 
  $sMyName );

  if ( ConnectionBroker::KillConnectionBrokerProcesses ( $sDMUser,
                                           $sInitFile ) ) {
    &logTrace ( 20032,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20033,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }

  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  ConnectionBroker::DeletePidFile ( $sResourceName );
  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------ 
Function : bool Kill ( IN String   CBName,
                       IN Integer  CleanReason,
                       IN String   DMBase,
                       IN String   DMUser,
                       IN String   EnvFile,
                       IN String   InitFile,
                       IN String   Password );

 Description : Forces shutdown of resource needed for clean entry point

 Parameters :

 $sCBName                   Documentum connection broker name managed by cluster server agent
 $iCleanReason              Why clean was called
 $sDMBase                   Absolute path of documentum connection broker binaries
 $sDMUser                   Unix user owner for Documentum connection broker
 $sEnvFile                  Absolute path to environment file to source
 $sInitFile                 The server initialization file for the Documentum connection broker
 $sPassword                 Encrypted password for teh documentum user.
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
------------------------------------------------------------------------------------------------------

=cut

sub Kill {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCBName,
       $iCleanReason,
       $sDMBase,
       $sDMUser,
       $sEnvFile,
       $sInitFile,
       $sPassword ) = @_;

  #-----------------------------
  # Figure out why were called..
  #-----------------------------
  my ( $sCleanReason, %hsCleanReasons ) = ();
  my @lsCleanStrings = qw ( Offline_timed_out    Offline_was_ineffective
                            Online_timed_out     Online_was_ineffective
                            Found_offline_unexpectedly Monitor_timed_out );

  @hsCleanReasons { 0 .. $#lsCleanStrings } = @lsCleanStrings; # Populate..
  ( $sCleanReason = $hsCleanReasons { $iCleanReason } ||       # Reason..
                    $iCleanReason ) =~ s/_/ /g;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20034, 
  "Subroutine <%s> called with:
  CBName                   [%s]
  CleanReason              [%s]
  DMBase                   [%s]
  DMUser                   [%s]
  EnvFile                  [%s]
  InitFile                 [%s]
  Password                 [%s]",
  $sMyName, $sCBName, $sCleanReason,  $sDMBase,
  $sDMUser, $sEnvFile, $sInitFile, $sPassword );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------------------------------------------------------------
  # Upper bound.. Time in EPOCH when entry pt started + VCS time limit..
  #---------------------------------------------------------------------
  my $iTillWhen = $^T + ConnectionBroker::GetEngineTimeout ( 1 ); # With margin..

  #--------------------------
  # Declare local variables..
  #--------------------------
  my ( $sCommand, $sCommandArgs, $iTimeout, $iExitCode,
       @lsFiles, @lsCmdout ) = ();
  
  #-------------------------------------
  # Define clean command and arguments..
  #-------------------------------------
  $sCommand     = "$sDMBase/dm_stop_$sCBName";
  $sCommandArgs = "-P$sPassword";

  unless ( $hsCleanReasons { $iCleanReason } eq 'Offline_timed_out' ||
           $hsCleanReasons { $iCleanReason } eq 'Offline_was_ineffective' ) {
    #-------------------
    # Try to be gentle..
    #-------------------
    &logInfo ( 20035,
    "<%s> Attempting to stop ConnectionBroker with Command [%s] Args [%s]",
    $sMyName, $sCommand, $sCommandArgs );
  
    #------------------------
    # Request shutdown..
    #------------------------
    if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
      &logError ( 20036,
      "<%s> Insufficient time to continue shutdown. Short by [%s] sec",
      $sMyName, abs ( $iTimeout ) );

      $iReturnCode = 0; # Failure..
    }
    elsif ( ConnectionBroker::ExecLocalCmd ( $sDMUser, 
                                      $sEnvFile, 
                                      $sCommand, 
                                      $sCommandArgs, 
                                      \$iExitCode, 
                                      $iTimeout, 
                                      \@lsCmdout, 
                                      $sPassword ) ) {
      &logTrace ( 20037, 
      "<%s> Cmd [%s] with args [%s] executed successfully",
      $sMyName, $sCommand, $sCommandArgs );
    }
  }

  #-----------------------------------------
  # Whack any persisting ConnectionBroker processes..
  #-----------------------------------------
  &logTrace ( 20038,
  "<%s> Killing ConnectionBroker processes for this instance", 
  $sMyName );

  if ( ConnectionBroker::KillConnectionBrokerProcesses ( $sDMUser,
                                           $sInitFile ) ) {
    &logTrace ( 20039,
    "<%s> Processes for this instance do not exist after shutdown",
    $sMyName );
  }
  else {
    &logError ( 20040,
    "<%s> Could not remove all processes for this instance. Exiting",
    $sMyName );

    $iReturnCode = 0; # Failure..
  }
  
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  ConnectionBroker::DeletePidFile ( $sResourceName );
  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : int GetResourceState ( IN String CBHost,
                                   IN String DMUser,
                                   IN String InitFile );

 Description : First level monitor to determine state of resource.

 Parameters :

 $sCBHost                   Hostname of virtual host for Documentum connection broker
 $sDMUser                   Unix user owner for Documentum connection broker
 $sInitFile                 The server initialization file for the Documentum connection broker
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50          VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
------------------------------------------------------------------------------------------------------

=cut

sub GetResourceState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCBHost,
       $sDMUser,
       $sInitFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20041, 
  "Subroutine <%s> called:
  CBHost                   [%s]
  DMUser                   [%s]
  InitFile                 [%s]",
  $sMyName, $sCBHost, $sDMUser, $sInitFile );
  
  my @liDummyPids = ();
  my $iReturnCode = ConnectionBroker::GetProcState ( \@liDummyPids,
                                              $sDMUser,
                                              $sInitFile,
                                              1 );  # Use FFLM..

  if ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    &logTrace ( 20042,
    "<%s> Found all required processes. Return code set to [%s]",
    $sMyName, $iReturnCode );
  }
  else {
    &logTrace ( 20043,
    "<%s> Did not find all required processes. Returning [%s]",
    $sMyName, $iReturnCode );
  }

  #--------------------------------
  # Populate the global PID array..
  #--------------------------------
  @liRunningProcs = @liDummyPids;
                                                                                
  &logTrace ( 20044,
  "<%s> Populating global running proclist with [%s] PIDs [%s]",
  $sMyName, scalar @liRunningProcs, join ( ' ', @liRunningProcs ) );

  return ( $iReturnCode );
}

=pod

------------------------------------------------------------------------------------------------------
 Function : int GetProcState ( OUT ListReference OnlinePids,
                               IN  String        DMUser,
                               IN  String        InitFile,
                               IN  Integer       UseFFLM );

 Description : First level monitor to determine state of resource.

 Parameters :

 $rliOnlinePids             PIDs of relevant running processes
 $sDMUser                   Unix user owner for Documentum connection broker
 $sInitFile                 The server initialization file for the Documentum connection broker
 $iUseFFLM                  Filtering procedure
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 105  ONLINE with confidence level 50          VCSDef_MonitorExit::PARTIAL_ONLINE
 110  ONLINE with confidence level 100         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
------------------------------------------------------------------------------------------------------

=cut

sub GetProcState {
  #----------------
  # Get Arguments..
  #----------------
  my ( $rliOnlinePids,
       $sDMUser,
       $sInitFile,
       $iUseFFLM ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20045, 
  "Subroutine <%s> called with:
  OnlinePids               [%s]
  DMUser                   [%s]
  InitFile                 [%s]
  UseFFLM                  [%s]",
  $sMyName, $rliOnlinePids, $sDMUser, $sInitFile, 
  $iUseFFLM );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #----------------------
  # Initialise Variables..
  #----------------------
  my ( $iProcessCount, $iUseRFLM, $iFound, $iUid, $iPid, $sLang,
       $sKey, $rhlProc, $sPidFile, $sProc, $sTemp,
       @hProcListHash, @lsUpdates, @lsValues,
       %hiProcPids, %hsSrvrDetails, %hsCopy ) = ();

  @$rliOnlinePids = ();

  #------------------------------------
  # Attempt to get monitoring details..
  #------------------------------------

  if ( $iUseFFLM != 2 ) { # Called for first level check..
    if ( ConnectionBroker::GetSrvrDetails ( $sInitFile, 
                                     \%hsSrvrDetails ) ) {
      &logTrace ( 20046,
      "<%s> Got details for monitoring ConnectionBroker instance", 
      $sMyName );
  
      %hsCopy   = %hsSrvrDetails;        # Backup..
      $iUseRFLM = scalar keys %hsCopy; # Set to entire count..
    }
    else {
      &logError ( 20047,
      "<%s> Failed to get details for monitoring ConnectionBroker instance",
      $sMyName );

      return VCSDef_MonitorExit::UNKNOWN;
    }
                                  
    #---------------------------------
    # Check if we attempt to be fast.. 
    #---------------------------------
    if ( $iUseFFLM ) { # FFLM..
      $iUseRFLM = 0;   # Hopefully skip..

      foreach $sKey ( keys %hsCopy ) {          # Use copy..
        @lsUpdates = ();                        # Reset..
        @lsValues  = @{ $hsCopy { $sKey } };     # Get values..
        $sTemp    .= "\nConnectionBroker process [$sKey]\n"; # Display string..

        #----------------
        # Read in pairs..
        #----------------
        while ( ( $sPidFile, $sProc ) = splice ( @lsValues, 0, 2 ) ) {

          #-----------
          # Run FFLM..
          #-----------
          $iFound = ConnectionBroker::FFLMStatus ( $sPidFile,
                                            $sProc,
                                            \$iPid );

          #----------------
          # Update status..
          #----------------
          push ( @lsUpdates, $sPidFile, $sProc, $iFound );

          $sTemp .= "PID File     [$sPidFile]\n";
          $sTemp .= "Srch Process [$sProc]\n";
          $sTemp .= "FFLM Status  [$iFound]\n\n";
        }

        @{ $hsCopy { $sKey } } = @lsUpdates;

        #------------------------
        # Handle monitor status..
        #------------------------
        if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
          push ( @$rliOnlinePids, $iPid ); # Push..
        }
        else {
          ++ $iUseRFLM; # Mark it for RFLM..
        }
      }

      #----------------
      # Print summary..
      #----------------
      &logTrace ( 20048,
      "<%s> ConnectionBroker instance process details ---%s\n\n",
      $sMyName, $sTemp );

      ( $sTemp, $sPidFile, $sProc ) = (); # Release memory..
    }
    else { # Use RLFM..
      #-----------------------------------------
      # Mark status of each service as offline..
      #-----------------------------------------
      $iFound = VCSDef_MonitorExit::OFFLINE; # Set to offline..
      foreach $sKey ( keys %hsCopy ) {
        @lsUpdates = (); # Reset..

        @lsValues = @{ $hsCopy { $sKey } }; # Get values..
        while ( scalar @lsValues ) {
          push ( @lsUpdates, splice ( @lsValues, 0, 2 ), $iFound );
        }

        @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
      }
    }
  }
  else { # Called for Kill..
    &logTrace ( 20049,
    "<%s> Called to get processes for User [%s] User [%s]",
    $sMyName, $sDMUser );
    ++ $iUseRFLM;
  }

  #-----------------------
  # Regular slow :-( FLM..
  #-----------------------
  if ( $iUseRFLM ) {
    &logTrace ( 20050,
    "<%s> Running regular first level monitor", 
    $sMyName );

    #----------------------------------
    # Set LANG for better performance..
    #----------------------------------
    $sLang = $ENV { LANG }; # Backup..
    $ENV { LANG } = 'C';    # Now set..

    #-----------------------------------
    # Call processes and store in hash..
    #-----------------------------------
    $iProcessCount = Proc::GetLongProcessListHash ( \@hProcListHash );

    #----------
    # Restore..
    #----------
    defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG }; # Restore..

    #----------------------
    # Check if we got any..
    #----------------------
    if ( $iProcessCount ) {
      &logTrace ( 20051,
      "<%s> Got [%s] processes from proc table for filtering",
      $sMyName, $iProcessCount );
    }
    else {
      &logError ( 20052,
      "<%s> Call to get process list did not return any processes",
      $sMyName );
      return VCSDef_MonitorExit::OFFLINE; # Offline..
    }

    #-----------------------
    # Search for processes..
    #-----------------------
    foreach $rhlProc ( @hProcListHash ) {
      #--------------------------------------------------------------
      # Each $rhlProc is a ref. to a hash with the following keys..
      # "UID"     =>  <user id>
      # "PID"     =>  <process id>
      # "PPID"    =>  <parent process id>
      # "COMMAND" =>  <command string>
      #--------------------------------------------------------------

      next unless defined $$rhlProc { UID }; # Bug in Proc::GetLongProcessListHash..
      next if $$rhlProc { PID } == $$;       # Skip self..
      $iPid = $$rhlProc { PID };             # Assign..

      &logTrace ( 20053, "<%s> Searching UID [%s] PID [%s] PPID [%s] CMD [%s]",
      $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
      $$rhlProc { COMMAND } );

      #-----------------------------------------------
      # Look for processes that were offline in FFLM..
      #-----------------------------------------------
      if ( $iUseFFLM != 2 ) { # Request to filter..
        foreach $sKey ( keys %hsCopy ) {
          @lsUpdates = (); # Reset..
          @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    
          #--------------------------------------------------
          # Should have details per three elements of array..
          # <PIDFILE> <PROCESS> <STATE>
          #--------------------------------------------------
          while ( ( $sPidFile, $sProc, $iFound ) =
                    splice ( @lsValues, 0, 3 ) ) {
            if ( VCSDef_MonitorExit::OFFLINE == $iFound ) { # Earlier marked offline..
              if ( $$rhlProc { COMMAND } =~ /$sProc/ ) {    # Now found..
                unshift ( @$rliOnlinePids, $iPid );         # Store..
  
                &logTrace ( 20054,
                "<%s> Found PID [%s] PPID [%s] CMD [%s] with filter [%s]",
                $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND },
                $sProc );

                $hiProcPids { $sKey } = $iPid;        # Store..
                $iFound = VCSDef_MonitorExit::ONLINE; # Mark online..
              }
            }
    
            #----------------
            # Update status..
            #----------------
            push ( @lsUpdates, $sPidFile, $sProc, $iFound );
          }
          @{ $hsCopy { $sKey } } = @lsUpdates; # Update..
        }
      }
      else { # Request for kill..
        &logTrace ( 20055,
        "<%s> Pushing PID [%s] with PPID [%s] Cmd [%s] on stack",
        $sMyName, $iPid, $$rhlProc { PPID }, $$rhlProc { COMMAND } );
        push ( @$rliOnlinePids, $iPid ); # Store..
      } 
    }
  }

  #------------------------------------
  # Set final status for each process..
  #------------------------------------
  my ( $iOnline, $iOffline, $sEntryPointName ) = ();

  ( $sEntryPointName = $0 ) =~ s#.*/##; # Basename..
  $sTemp = ''; # Ensure..

  foreach $sKey ( keys %hsCopy ) {
    @lsValues = @{ $hsCopy { $sKey } }; # Get values..
    $sTemp .= "\nProcess      [$sKey]\n";

    while ( ( $sPidFile, $sProc, $iFound ) =
              splice ( @lsValues, 0, 3 ) ) {
      #-----------------------------------
      # Create PID files for use in FFLM..
      #-----------------------------------
      if ( $iUseRFLM                                 &&      # In RFLM..
           $iUseFFLM != 2                            &&      # Asked to filter..
           $hiProcPids { $sKey }                     &&      # PID populated..
           ( $sEntryPointName =~ /^(monitor|online)$/o ) ) { # In monitor or online..

        &logTrace ( 20056,
        "<%s> Creating PID file [%s] with PID [%s]",
        $sMyName, $sPidFile, $hiProcPids { $sKey } );

        #---------
        # Create..
        #---------
        if ( ConnectionBroker::CreatePidFile ( $sPidFile,
                                        $hiProcPids { $sKey } ) ) {
          &logTrace ( 20057,
          "<%s> Successfully created PID file [%s]", 
          $sMyName, $sPidFile );
        }
        else {
          &logError ( 20058,
          "<%s> Could not create PID file [%s] successfully",
          $sMyName, $sPidFile );
        }
      }

      #-------------
      # Set status..
      #-------------
      if ( VCSDef_MonitorExit::ONLINE == $iFound ) {
        $iOnline = $iFound;
      }
      elsif ( VCSDef_MonitorExit::OFFLINE == $iFound ) {
        $iOffline = $iFound;
      }

      $sTemp .= "Srch Process [$sProc]\n";
      $sTemp .= "FLM Status   [$iFound]\n\n";
    }

    $sTemp .= "--------------------";
  }

  #-------------------
  # Set final status..
  #-------------------
  $iReturnCode = $iOffline || $iOnline || VCSDef_MonitorExit::UNKNOWN;

  #------------------------------------------------
  # Reset status based on PIDs, if not in monitor..
  #------------------------------------------------
  if ( $sEntryPointName ne 'monitor'                   &&
       ( VCSDef_MonitorExit::OFFLINE == $iReturnCode ) &&
       scalar @$rliOnlinePids ) {
    $iReturnCode = VCSDef_MonitorExit::PARTIAL_ONLINE;

    &logTrace ( 20059,
    "<%s> Found [%s] PIDs [%s] for this instance of ConnectionBroker.
    Resetting return code to [%s]",
    $sMyName, scalar @$rliOnlinePids, join ( ' ', @$rliOnlinePids ),
    $iReturnCode );
  }

  #-----------------------
  # Report on FLM status..
  #-----------------------
  if ( $iUseFFLM != 2 ) { # Request to filter..
    &logTrace ( 20060,
    "<%s> ConnectionBroker process details ---\n%s\nReturn code is [%s]",
    $sMyName, $sTemp, $iReturnCode );
    $sTemp = undef; # Release..
  }

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : int Monitor ( IN String    CBHost,
                          IN Integer   CBPort,
                          IN String    DMHome,
                          IN String    DMUser,
                          IN String    EnvFile,
                          IN String    MonitorProgram,
                          IN Integer   SecondLevelMonitor );

 Description: Performs the in-depth monitoring

 Parameters :

 $sCBHost                   Hostname of virtual host for Documentum connection broker
 $iCBPort                   The port on which the Documentum connection broker listens
 $sDMHome                   Absolute path of documentum product installation
 $sDMUser                   Unix user owner for Documentum connection broker
 $sEnvFile                  Absolute path to environment file to source
 $sMonitorProgram           Full path to external monitor program
 $iSecondLevelMonitor       Cycle count to run in-depth monitoring
 
 Returns : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if in-depth monitoring cannot be completed.
------------------------------------------------------------------------------------------------------

=cut

sub Monitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCBHost,
       $iCBPort,
       $sDMHome,
       $sDMUser,
       $sEnvFile,
       $sMonitorProgram,
       $iSecondLevelMonitor ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20061, 
  "Subroutine <%s> called with:
  CBHost                   [%s]
  CBPort                   [%s]
  DMHome                   [%s]
  DMUser                   [%s]
  EnvFile                  [%s]
  MonitorProgram           [%s]
  SecondLevelMonitor       [%s]",
  $sMyName, $sCBHost, $iCBPort, $sDMHome, $sDMUser, 
  $sEnvFile, $sMonitorProgram, $iSecondLevelMonitor );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  my $iTillWhen = undef;

  #----------------------------------------------
  # In-depth monitoring, if correct time to run..
  #----------------------------------------------
  if ( VCSagentFW::RunSecondLevelMonitor ( $iSecondLevelMonitor ) ) {
    #--------------------------
    # Just In Time Validation..
    #--------------------------
    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         ConnectionBroker::DelayedAttributeCheck ( $sCBHost,
                                            $iCBPort,
                                            $sDMHome,
                                            $sDMUser,
                                            $sEnvFile ) ) ) {
      &logTrace ( 20062,
      "<%s> Delayed attribute check completed successfully",
      $sMyName );
    }
    else {
      &logError ( 20063,
      "<%s> Attribute validation failed. Returning [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }

    &logTrace ( 20064,
    "<%s> Proceeding with second level monitoring", 
    $sMyName );

    #-------------------------------------------------------
    # Upper bound..
    # Time in EPOCH when entry pt started + VCS time limit..
    #-------------------------------------------------------
    $iTillWhen = $^T + ConnectionBroker::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE == ( $iReturnCode =
         ConnectionBroker::SecondLevelMonitor ( $sCBHost,
                                         $iCBPort,
                                         $sDMHome,
                                         $sDMUser,
                                         $sEnvFile,
                                         $iTillWhen ) ) ) {

      &logTrace ( 20065,
      "<%s> ConnectionBroker instance online with status [%s] after SLM",
      $sMyName, $iReturnCode );
    }
    else {
      &logError ( 20066,
      "<%s> ConnectionBroker instance NOT online after SLM. Status [%s]",
      $sMyName, $iReturnCode );
      return $iReturnCode;
    }
  }

  #------------------------------------------------
  # Exit if external monitor has not been defined..
  #------------------------------------------------
  return $iReturnCode unless defined $sMonitorProgram;

  #---------------------------------------
  # MonitorProgram defined..
  # Separate the executable and the args..
  #---------------------------------------
  my ( $sExtCmd, $sExtCmdArgs ) = $sMonitorProgram # Separate
                                =~ /(\S+)\s*(.*)/; # Cmd and args..

  $sExtCmdArgs = '' unless defined $sExtCmdArgs;   # Prevent undefs..

  if ( $sExtCmd ) {                                # Defined..
    #----------------------------------------
    # Get time bearings, if not got earlier..
    #----------------------------------------
    $iTillWhen ||= $^T + ConnectionBroker::GetEngineTimeout ( 1 );

    if ( VCSDef_MonitorExit::ONLINE ==             # and found online..
         ( $iReturnCode =                          # by the status..
         ConnectionBroker::ExternalMonitor ( $sDMUser,
                                      $sEnvFile,
                                      $sExtCmd,
                                      $sExtCmdArgs,
                                      $iTillWhen ) ) ) {
      &logTrace ( 20067,
      "<%s> ConnectionBroker instance found online with status [%s]
      after executing external monitor [%s] with args [%s]",
      $sMyName, $iReturnCode, $sExtCmd, $sExtCmdArgs );
    }
    else {
      &logError ( 20068,
      "<%s> ConnectionBroker instance NOT online after executing external 
      monitor [%s] with args [%s]. Status [%s]",
      $sMyName, $sExtCmd, $sExtCmdArgs, $iReturnCode );
    }
  }

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : int DelayedAttributeCheck ( IN String   CBHost,
                                        IN Integer  CBPort,
                                        IN String   DMHome,
                                        IN String   DMUser,
                                        IN String   EnvFile );

 Description : Just In Time Validation

 Parameters :

 $sCBHost                   Hostname of virtual host for Documentum connection broker
 $iCBPort                   The port on which the Documentum connection broker listens
 $sDMHome                   Absolute path of documentum product installation
 $sDMUser                   Unix user owner for Documentum connection broker
 $sEnvFile                  Absolute path to environment file to source
 
 Returns  : Integer value as defined below:

 110  ONLINE with confidence level 110          VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined     VCSDef_MonitorExit::UNKNOWN
------------------------------------------------------------------------------------------------------

=cut

sub DelayedAttributeCheck {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCBHost,
       $iCBPort,
       $sDMHome,
       $sDMUser,
       $sEnvFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20069, 
  "Subroutine <%s> called with:
  CBHost                   [%s]
  CBPort                   [%s]
  DMHome                   [%s]
  DMUser                   [%s]
  EnvFile                  [%s]",
  $sMyName, $sCBHost, $iCBPort, $sDMHome, $sDMUser, $sEnvFile );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..

  #--------------------------------------------
  # Create hash to weed out empty args values..
  #--------------------------------------------
  my @lsArgs = qw ( CBHost CBPort DMHome DMUser EnvFile );
  my @lsReqdArgs = qw ( CBHost CBPort DMHome DMUser EnvFile );
  my ( @lsMissing, %hsArgList ) = ();

  @hsArgList { @lsArgs } = @_; # Assign..

  #-----------------------
  # Check what's missing..
  #-----------------------
  if ( @lsMissing = grep { $hsArgList { $_ } !~ /\S/o } @lsReqdArgs ) {
    &logError ( 20070,
    "<%s> The following mandatory fields were not set [%s], Total [%s]",
    $sMyName, join ( ' ', @lsMissing ), scalar @lsMissing );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #--------------------
  # Return, if failed..
  #--------------------
  if ( VCSDef_MonitorExit::ONLINE != $iReturnCode ) {
    return $iReturnCode;
  }

  #------------------------------
  # Validate parameter EnvFile..
  #------------------------------
  if ( ! -e $sEnvFile ) {
    &logError ( 20071,
    "<%s> EnvFile [%s] file does not exist",
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  if ( ! -e $sDMHome ) {
    &logError ( 20072,
    "<%s> DMHome [%s] does not exist",
    $sMyName, $sDMHome );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }

  #-----------------------------
  # Validate parameter DMUser..
  #-----------------------------

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : int SecondLevelMonitor ( IN String   CBHost,
                                     IN Integer  CBPort,
                                     IN String   DMHome,
                                     IN String   DMUser,
                                     IN String   EnvFile,
                                     IN Integer  TillWhen );

 Description : Does in-depth monitoring, if enabled.

 Parameters :

 $sCBHost                   Hostname of virtual host for Documentum connection broker
 $iCBPort                   The port on which the Documentum connection broker listens
 $sDMHome                   Absolute path of documentum product installation
 $sDMUser                   Unix user owner for Documentum connection broker
 $sEnvFile                  Absolute path to environment file to source
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
------------------------------------------------------------------------------------------------------

=cut

sub SecondLevelMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sCBHost,
       $iCBPort,
       $sDMHome,
       $sDMUser,
       $sEnvFile,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20073, 
  "Subroutine <%s> called with:
  CBHost                   [%s]
  CBPort                   [%s]
  DMHome                   [%s]
  DMUser                   [%s]
  EnvFile                  [%s]
  TillWhen                 [%s]",
  $sMyName, $sCBHost, $iCBPort, $sDMHome, $sDMUser, 
  $sEnvFile, $iTillWhen );  

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iTimeout, $iExitCode, $sCommand, $sCommandArgs, @lsCmdout ) = ();
  
  #------------------------------------
  # Check if port can be connected to..
  #------------------------------------
  if ( Comms::CheckPort ( $sCBHost,
                          $iCBPort ) ) {
    &logTrace ( 20074,
    "<%s> Server listening on Host [%s] Port [%s]: Socket method",
    $sMyName, $sCBHost, $iCBPort );
  }
  else {
    &logError ( 20075,
    "<%s> Server NOT listening on Host [%s] Port [%s]: Socket method",
    $sMyName, $sCBHost, $iCBPort );

    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Return, unless online..
  #------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  #---------------------------------------------
  # Declare monitor command and args variables..
  #---------------------------------------------
  $sCommand = "$sDMHome/bin/dmqdocbroker";
  $sCommandArgs = "-t $sCBHost -p $iCBPort -c ping";

  #-------------------------------
  # Validate and execute command..
  #-------------------------------
  if ( ( $iTimeout = $iTillWhen - time ) <= 10 ) { # Get time bearings..
    &logError ( 20076,
    "<%s> Insufficient time to continue in-depth monitoring.
    Short by [%s] sec",
    $sMyName, abs ( $iTimeout ) );

    $iReturnCode = VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( ConnectionBroker::ExecLocalCmd ( $sDMUser, 
                                    $sEnvFile, 
                                    $sCommand, 
                                    $sCommandArgs, 
                                    \$iExitCode, 
                                    $iTimeout, 
                                    \@lsCmdout ) ) {
    &logTrace ( 20077,
    "<%s> Cmd [%s] with args [%s] executed successfully",
    $sMyName, $sCommand, $sCommandArgs );
    
  }
  elsif ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iReturnCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                               VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  else {
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : int ExternalMonitor ( IN String  DMUser,
                                  IN String  EnvFile,
                                  IN String  Cmd,
                                  IN String  CmdArgs,
                                  IN Integer TillWhen );

 Description : Handles user defined external monitor program

 Parameters :

 $sDMUser                   Unix user owner for Documentum connection broker
 $sEnvFile                  Absolute path to environment file to source
 $sCmd                      User defined utilty to execute
 $sCmdArgs                  Command line arguments
 $iTillWhen                 Time in EPOCH left for IInd level monitoring
 
 Returns  : Integer value as defined below:

 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 110         VCSDef_MonitorExit::ONLINE
 99   UNKNOWN; State couldn't be determined    VCSDef_MonitorExit::UNKNOWN
 ??   HUNG: External monitor timed out         VCSDef_MonitorExit::TIMEOUT

 The entry point hangs if the command cannot be completed.
------------------------------------------------------------------------------------------------------

=cut

sub ExternalMonitor {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDMUser,
       $sEnvFile,
       $sCmd,
       $sCmdArgs,
       $iTillWhen ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20078, 
  "Subroutine <%s> called with:
  DMUser                   [%s]
  EnvFile                  [%s]
  Cmd                      [%s]
  CmdArgs                  [%s]
  TillWhen                 [%s]",
  $sMyName, $sDMUser, $sEnvFile, $sCmd, 
  $sCmdArgs, $iTillWhen );

  #------------------
  # Hopeful success..
  #------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE;

  #------------------------
  # Setup local variables..
  #------------------------
  my ( $iExitCode, $iTimeout, @lsCmdout ) = ();
  my $iBuf = 5; # Secs..

  #-------------------------
  # Handled exit code hash..
  #-------------------------
  my %hiExitCodes = (  0, VCSDef_MonitorExit::ONLINE,
                     110, VCSDef_MonitorExit::ONLINE,
                       1, VCSDef_MonitorExit::OFFLINE,
                     100, VCSDef_MonitorExit::OFFLINE,
                          VCSDef_MonitorExit::TIMEOUT,
                          VCSDef_MonitorExit::TIMEOUT );

  #---------------------------------------------
  # Check if command can be found and executed..
  #---------------------------------------------
  if ( ! -e $sCmd ) {
   &logError ( 20079,
    "<%s> External monitor [%s] does not exist",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ! -x $sCmd ) {
   &logError ( 20080,
    "<%s> External monitor [%s] not executable.",
    $sMyName, $sCmd );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( $sEnvFile =~ /\S/ && ! -e $sEnvFile ) {
    &logError ( 20081,
    "<%s> EnvFile [%s] does not exist", 
    $sMyName, $sEnvFile );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }
  elsif ( ( $iTimeout = $iTillWhen - ( time + $iBuf ) ) <= 5 ) { # Lack of time..
    &logError ( 20082,
    "<%s> Insufficient time to proceed with external monitor [%s].
    Short by [%s] sec", $sMyName, $sCmd, abs ( $iTimeout ) );
    $iReturnCode = VCSDef_MonitorExit::UNKNOWN;
  }

  #--------------------------------------------
  # Return if validation failed at this point..
  #--------------------------------------------
  unless ( VCSDef_MonitorExit::ONLINE == $iReturnCode ) {
    return $iReturnCode;
  }

  &logTrace ( 20083,
  "<%s> Executing external monitor with Command [%s] Args [%s] in [%s] secs",
  $sMyName, $sCmd, $sCmdArgs, $iTimeout );

  #----------------------
  # Execute the command..
  #----------------------
  ConnectionBroker::ExecLocalCmd ( $sDMUser, 
                            $sEnvFile, 
                            $sCmd, 
                            $sCmdArgs, 
                            \$iExitCode, 
                            $iTimeout, 
                            \@lsCmdout );
                               
  #------------------------
  # Report on exit status..
  #------------------------
  if ( ! defined $iExitCode ) {
    $lsCmdout [0] ||= 'undef'; # Set if not defined..
    $iExitCode = $lsCmdout [0] eq 'timeout'? VCSDef_MonitorExit::TIMEOUT: # Timeout..
                                             VCSDef_MonitorExit::UNKNOWN; # Failure..
  }
  elsif ( $hiExitCodes { $iExitCode } &&
          VCSDef_MonitorExit::ONLINE ==
          $hiExitCodes { $iExitCode } ) {
    &logTrace ( 20084,
    "<%s> Command [%s] with args [%s] executed successfully with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }
  else {
    &logWarn ( 20085,
    "<%s> Command [%s] with args [%s] execution failed with exit code [%s].
    Cmd output [%s]",
    $sMyName, $sCmd, $sCmdArgs, $iExitCode, join ( chr 012, @lsCmdout ) );
  }

  #--------------------------------
  # Setup appropriate return code..
  #--------------------------------
  $iReturnCode = $hiExitCodes { $iExitCode } ||
                 VCSDef_MonitorExit::UNKNOWN;

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : bool CreatePidFile ( IN String   PidFile,
                                 In Integer  Pid );


 Description : Creates PID file with given PID

 Parameters :

 $sPidFile     File to create
 $iPid         PID to store

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
------------------------------------------------------------------------------------------------------

=cut

sub CreatePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $iPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20086, 
  "Subroutine <%s> called with:
  PidFile         [%s]
  Pid             [%s]",
  $sMyName, $sPidFile, $iPid );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #-----------------------
  # Check against undefs..
  #-----------------------
  unless ( defined $iPid && $iPid =~ /^\d+$/o) {
    $iPid ||= 'undef'; # Prevent undefs.
    &logError ( 20087, 
    "<%s> PID [%s] is not a valid positive number",
    $sMyName, $iPid );
    return 0; # Proceed no further..
  }

  #------------------
  # Create PID file..
  #------------------
  &logTrace ( 20088,
  "<%s> Creating PID file [%s] with PID [%s]",
  $sMyName, $sPidFile, $iPid );

  if ( sysopen ( PIDFILE, $sPidFile, O_WRONLY|O_SYNC|O_CREAT|O_TRUNC,
                 0600 ) ) {
    print PIDFILE $iPid, "\n"; # Update..

    unless ( close ( PIDFILE ) ) {
      &logWarn ( 20089,
      "<%s> Failed to close PidFile [%s] after update: [%s]",
      $sMyName, $sPidFile, $! );
      $iReturnCode = 0;
    }
  }
  else {
    &logWarn ( 20090,
    "<%s> Failed to create PID file [%s] with PID [%s]",
    $sMyName, $sPidFile, $iPid );
    $iReturnCode = 0;
  }

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------
 Function : bool DeletePidFile ( IN String ResourceName );


 Description : Deletes PID files for the VCS resource.

 Parameters :

 $sResourceName ResourceName for which the PID files needs to be deleted.

 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
------------------------------------------------------------------------------------------------

=cut

sub DeletePidFile {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sResourceName ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20091, 
  "Subroutine <%s> called with:
  ResourceName        [%s]",
  $sMyName, $sResourceName );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  my ( $sPidFileDir, $iRet ) = ();
  
  #-------------------
  # Delete PID files..
  #-------------------
  &logTrace ( 20092,
  "<%s> Deleting PID files for VCS resource [%s]",
  $sMyName, $sResourceName );

  $sPidFileDir = "/tmp/.VRTSConnectionBroker/" . $sResourceName . "/PID" ;

  if ( -d $sPidFileDir ) {
    $iRet = system("/bin/rm -rf $sPidFileDir") ;

    if ( $iRet == 0 ) {
      &logTrace ( 20093,
      "<%s> Successfully deleted PID files in the directory [%s].",
      $sMyName, $sPidFileDir );
      $iReturnCode = 0;
    }
    else {
      &logWarn ( 20094,
      "<%s> Failed to delete PID files in directory [%s]. Remove the files manually.",
      $sMyName, $sPidFileDir );
      $iReturnCode = 1;
    }
  }
  else {
    &logTrace ( 20095,
    "<%s> PID files for VCS resource [%s] doesn't exist.",
    $sMyName, $sResourceName );
  }

  return $iReturnCode;
} 

=pod

------------------------------------------------------------------------------------------------------
 Function : int FFLMStatus ( IN String  PidFile,
                             IN String  Proc,
                             OUT RefInt Pid );

 Description: Checks status of Process listed in PidFile, if running

 Parameters :

 $sPidFile       Absolute path to PID file of service
 $sProc          Process string to search for in proc table
 $riPid          Reference to existing PID

 Returns : Integer value as defined below:
 100  OFFLINE                                  VCSDef_MonitorExit::OFFLINE
 110  ONLINE with confidence level 100         VCSDef_MonitorExit::ONLINE
------------------------------------------------------------------------------------------------------

=cut

sub FFLMStatus {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sPidFile,
       $sProc,
       $riPid ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20096, 
  "Subroutine <%s> called with:
  PidFile                [%s]
  Proc                   [%s]
  riPid                  [%s]",
  $sMyName, $sPidFile, $sProc, $riPid );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = VCSDef_MonitorExit::ONLINE; # Success..
  $$riPid = 0; # Reset..

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $iPid, $rhlProc, @lProcessHash ) = ();

  #----------------------------------
  # Start by testing the attributes..
  #----------------------------------
  if ( $sProc !~ /\S/ ) {
    &logError ( 20097, 
    "<%s> Called with empty [%s] process string",
    $sMyName, $sProc );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -e $sPidFile ) {
    &logTrace ( 20098, 
    "<%s> PID file [%s] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }
  elsif ( ! -s $sPidFile ) {
    &logWarn ( 20099, 
    "<%s> PID file [%s] is of zero bytes",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #---------------------
  # Return, if offline..
  #---------------------
  return $iReturnCode unless VCSDef_MonitorExit::ONLINE == $iReturnCode;
 
  if ( open ( PIDFILE, $sPidFile ) ) {
    ( $iPid ) = grep { ( s/\n// || 1 ) && # Chomp..
                         s/.*?(\d+)$/$1/  # Get PID..
                } <PIDFILE>;

    $iPid = '' unless defined $iPid;      # Avoid undefs..

    close ( PIDFILE ) ||                  # Close file..
    &logWarn ( 20100,
    "<%s> Could not close PID file [%s] after read [%s]",
    $sMyName, $sPidFile, $! );
  }
  else {
    &logError ( 20101,
    "<%s> Failed to open PID file [%s] for read: [%s]",
    $sMyName, $sPidFile, $! );
    return VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  #------------------------
  # Check if PID is valid..
  #------------------------
  if ( $iPid =~ /\d+/ &&     # If the valid number..
       kill ( 0, $iPid ) ) { # is a PID that's alive..

    &logTrace ( 20102,
    "<%s> PID [%s] from PID file [%s] exists",
    $sMyName, $iPid, $sPidFile );

    #-----------------------
    # Search in proc table..
    #-----------------------
    if ( ConnectionBroker::GetPidDetails ( $iPid,
                                    \@lProcessHash ) ) {
      #--------------
      # Got details..
      #--------------
      foreach $rhlProc ( @lProcessHash ) {
        &logTrace ( 20103,
        "<%s> Process details: UID [%s] PID [%s] PPID [%s] ARGS [%s]",
        $sMyName, $$rhlProc { UID }, $$rhlProc { PID }, $$rhlProc { PPID },
        $$rhlProc { COMMAND } );

        #--------------
        # Now compare..
        #--------------
        if ( $$rhlProc { COMMAND } =~ /$sProc/ ) { # Filter match..
          &logTrace ( 20104,
          "<%s> PID [%s] owned by UID [%s] with ARGS [%s] contains [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $$riPid = $iPid; # Store..
        }
        else {
          &logTrace ( 20105,
          "<%s> Ignoring PID [%s] owned by UID [%s] with ARGS [%s] with string [%s]",
          $sMyName, $iPid, $$rhlProc { UID }, $$rhlProc { COMMAND }, $sProc );
          $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
        }
      }
    }
    else {
      &logTrace ( 20106,
      "<%s> Could not get details for PID [%s]",
      $sMyName, $iPid );
      $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
    }
  }
  else {
    &logTrace ( 20107, "<%s> PID [$iPid] does not exist",
    $sMyName, $sPidFile );
    $iReturnCode = VCSDef_MonitorExit::OFFLINE; # Failure..
  }

  @lProcessHash = (); # Truncate..
  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : bool KillConnectionBrokerProcesses ( IN String DMUser,
                                          IN String InitFile );

 Description  : Kills processes for this instance of ConnectionBroker
 
 Parameters :

 $sDMUser                 Unix user owner for Documentum connection broker
 $sInitFile               The server initialization file for the Documentum connection broker
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
------------------------------------------------------------------------------------------------------

=cut

sub KillConnectionBrokerProcesses {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sDMUser,
       $sInitFile ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20108, 
  "Subroutine <%s> called with:
  DMUser                   [%s]
  InitFile                 [%s]",
  $sMyName, $sDMUser, $sInitFile );
 
  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1; # Success..

  #-------------------------
  # Set up local variables..
  #-------------------------
  my ( @liGoners, @liUnDead ) = ();

  #---------------------------------------------
  # Get list of processes running for this Uid..
  #---------------------------------------------
  if ( ( caller 1 )[3] eq 'ConnectionBroker::Start' ) { # During online..
    @liGoners = @liRunningProcs;              # Get PIDs to kill..
  }
  else {
    ConnectionBroker::GetProcState ( \@liGoners,
                              $sDMUser,
                              $sInitFile,
                              0 );  # Set filtering choice..
  }

  #--------------------------
  # Sort in ascending order..
  #--------------------------
  @liGoners = sort { $a <=> $b } @liGoners;

  #------------------------------------------------------------------
  # We do a two level kill. The SIGINQ ensures  that  the  PID  is  a
  # valid process before we actually issue the SIGKILL. We don't want
  # to report a failure for SIGKILL in case  of  defunct  or  missing
  # processes.
  #------------------------------------------------------------------
  if ( scalar @liGoners ) {
    #----------------------
    # Failed to kill some..
    #----------------------
    if ( @liUnDead = grep {
                         kill ( 0, $_ ) && # Valid..
                       ! kill ( 9, $_ )    # and could NOT be killed..
                     } @liGoners ) {       # From what's found..
      &logError ( 20109,
      "<%s> PIDs not killed [%s] Total not killed [%s]",
      $sMyName, join ( ' ', @liUnDead ), scalar @liUnDead );
      $iReturnCode = 0; # Failure..
    }
    else { # All killed..
      &logInfo ( 20110,
      "<%s> Killed PIDs [%s] Total killed: [%s]",
      $sMyName, join ( ' ', @liGoners ), scalar @liGoners );
    }
  }
  else {
    &logTrace ( 20111,
    "<%s> Could not locate any process this instance of ConnectionBroker",
    $sMyName );
  }

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : bool ExecLocalCmd ( IN String         User,
                                IN String         EnvFile,
                                IN String         Command,
                                IN String         CommandArgs,
                                OUT RefInt        ExitCode,
                                IN Integer        Timeout,
                                OUT ListReference Cmdout 
                                IN String         HideString );

 Description : Wrapper function for
               Sys::RunWithEnvCmdWithOutputWithTimeOut to collect output
               of the command.
 
 Parameters :

 $sUser            UNIX user to run the command as
 $sEnvFile         Absolute path to location of EnvFile to source
 $sCommand         Absolute path to location of command
 $sCommandArgs     Arguments to command
 $riExitCode       Reference to collect exit code of command
 $iTimeout         Time limit to execute the command
 $rlsCmdout        Stores o/p from command if successful
 $sHideString      String in CommandArgs to hide
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
------------------------------------------------------------------------------------------------------

=cut

sub ExecLocalCmd {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sUser,
       $sEnvFile,
       $sCommand,
       $sCommandArgs,
       $riExitCode,
       $iTimeout,
       $rlsCmdout,
       $sHideString  ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20112, 
  "Subroutine <%s> called with:
  User        [%s]
  EnvFile     [%s]
  Command     [%s]
  CommandArgs [%s]
  riExitCode  [%s]
  Timeout     [%s]
  rlsCmdout   [%s]",
  $sMyName, $sUser, $sEnvFile, $sCommand, 
  $sCommandArgs, $riExitCode, $iTimeout, $rlsCmdout );

  #---------------------------------------
  # Setup return code to hopeful success..
  #---------------------------------------
  my $iReturnCode = 1;
  my ( $sDeCrPasswd, $sCmdDeCr, $iExitCode ) = ();
  my $sCommandArgsDecr = $sCommandArgs; 
  #-------------------------
  # Get decrypted password..
  #-------------------------
  if ( ( defined $sHideString ) && ( $sHideString =~ /\S+/ ) ) {
    #-----------
    # Validate..
    #-----------
    if ( ! -x $sCmdVcsDecrypt ) {
      &logError ( 20113,
      "<%s> Command [%s] is not executable",
      $sMyName, $sCmdVcsDecrypt );
      return 0; # Failure..
    }

    $sCmdDeCr = $sCmdVcsDecrypt;                          # Get global..
    ( $sDeCrPasswd ) = grep {
                         s/([^\w])/\\$1/g || 1
                       } readpipe ( "$sCmdDeCr $sHideString" );

    #--------------------
    # Collect exit code..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logError ( 20114, "<%s> Failed to decrypt crypted string [%s]",
      $sMyName, $sHideString );
      return 0; # Failure..
    }

    $sCommandArgsDecr =~ s/$sHideString/$sDeCrPasswd/g;
  }

  #-------------------
  # Set up arguments..
  #-------------------
  my $bSourceUserEnv = 0; # Do not source..
  my $bTimedOut = 0;      # Timeout variable..

  @$rlsCmdout = (); # Reset..

  #-----------
  # Validate..
  #-----------
  if ( ! -e $sCommand ) {
    &logError ( 20115,
    "<%s> Cmd [%s] does not exist", 
    $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( ! -x $sCommand ) {
    &logError ( 20116,
    "<%s> Cmd [%s] is not executable", 
    $sMyName, $sCommand );
    $iReturnCode = 0; # Failure..
  }
  elsif ( $iTimeout <= 1 ) {
    &logError ( 20117,
    "<%s> Insufficient time [%s] to execute Cmd [%s] with Args [%s].",
    $sMyName, $iTimeout, $sCommand, $sCommandArgs );
    $iReturnCode = 0; # Failure..
  }

  $iReturnCode || return $iReturnCode; # Return if failed.. 

  #----------------------
  # Set LOGNAME to user..
  #----------------------
  my $sOldLogName = $ENV { LOGNAME }; # Backup..
  $ENV { LOGNAME } = $sUser;          # Now set..

  &logTrace ( 20118,
  "<%s> Executing Command [%s] with Args [%s]",
  $sMyName, $sCommand, $sCommandArgs );

  $iReturnCode = Sys::RunWithEnvCmdWithOutputWithTimeOut ( $sEnvFile,
                                                           $bSourceUserEnv,
                                                           $sCommand,
                                                           $sCommandArgsDecr,
                                                           $riExitCode,
                                                           $sUser,
                                                           $rlsCmdout,
                                                           \$bTimedOut,
                                                           $iTimeout,
                                                           undef,
                                                           undef,
                                                           $sDeCrPasswd );

  #---------------------------
  # Remove trailing newlines..
  #---------------------------
  chomp @$rlsCmdout;

  #--------------------
  # Check return code..
  #--------------------
  if ( $iReturnCode ) {   # No error with command execution..
    if ( $$riExitCode ) { # Failure..
      &logTrace ( 20119,
      "<%s> Command [%s] with args [%s] returned exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
      $iReturnCode = 0; # Failure..
    }
    else {
      &logTrace ( 20120,
      "<%s> Command [%s] with args [%s] executed successfully with exit code [%s]",
      $sMyName, $sCommand, $sCommandArgs, $$riExitCode );
    }
  }
  else { # Execution error..
    @$rlsCmdout = (); # Reset..

    if ( $bTimedOut ) {  # We timed out..
      &logError ( 20121,
      "<%s> Command [%s] with args [%s] did not complete within [%s] seconds",
      $sMyName, $sCommand, $sCommandArgs, $iTimeout );
      $$rlsCmdout [0] = 'timeout'; # Timeout..
    }
    else { # Failure due to something else..
      &logError ( 20122,
      "<%s> Failed in call to RunWithEnvCmdWithOutputWithTimeOut",
      $sMyName );
    }
  }

  #----------
  # Restore..
  #----------
  defined $sOldLogName ? $ENV { LOGNAME } = $sOldLogName : delete $ENV { LOGNAME };

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : bool GetPidDetails ( IN Integer  Pid,
                                 OUT RefList ProcList );


 Description : Gets long process details for a known Pid if running

 Parameters :

 $iPid         PID whose details are requested
 $rlProcList   Ref to List to store details
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
------------------------------------------------------------------------------------------------------

=cut

sub GetPidDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $iPid,
       $rlProcList ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20123, 
  "Subroutine <%s> called with:
  Pid             [%s]
  rlProcList      [%s]",
  $sMyName, $iPid, $rlProcList );

  #--------------------------------------
  # Set Return Code Optimistically to 1..
  #--------------------------------------
  my $iReturnCode = 1;

  #---------------
  # Validate PID..
  #---------------
  unless ( defined $iPid    &&   # Catch undefs..
           $iPid =~ /^\d+$/ &&   # And invalid PIDs..
           kill ( 0, $iPid ) ) { # And defunct ones..
    &logError ( 20124,
    "<%s> PID [%s] is invalid", 
    $sMyName, $iPid );
    return 0;
  }

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sPsCmd, $sPsCmdArgs, $sLang, $sFmt, $sUnix95, 
       $sSwitch, $iExitCode, %hsProc ) = ();

  $sSwitch = $^O;    # Get OS..
  @$rlProcList = (); # Truncate..

  #-------------
  # Store LANG..
  #-------------
  $sLang = $ENV { LANG };
  $ENV { LANG } = 'C';

  #---------------------------
  # Switch to the correct OS..
  #---------------------------

  if ( $sSwitch eq "solaris" ) {
      $sPsCmd = "/usr/ucb/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt = '[\da-z]+\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  } elsif ( $sSwitch eq "hpux" ) {
      #-----------------------
      # Force XPG4 behaviour..
      #-----------------------
      $sUnix95 = $ENV { UNIX95 };
      $ENV { UNIX95 } = 'XPG4';

      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "-p $iPid -xxlo uid,pid,ppid,args";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  } elsif ( $sSwitch eq "aix" ) {
      $sPsCmd = "/usr/bin/ps";
      $sPsCmdArgs = "axwwl $iPid";
      $sFmt =  '[\da-e]+\s+[A-Z]\s+([\d]+)\s+([\d]+)\s+([\d]+)\s+.*?[\d]:[\d]{2}\s(.*)';
  } elsif ( $sSwitch eq "linux" ) {
      $sPsCmd = "/bin/ps";
      $sPsCmdArgs = "wwo uid,pid,ppid,args $iPid";
      $sFmt = '([\d]+)\s+([\d]+)\s+([\d]+)\s+(.*)';
  }

  &logTrace ( 20125, 
  "<%s> OS [%s] PsCmd [%s] PsCmdArgs [%s]",
  $sMyName, $sSwitch, $sPsCmd, $sPsCmdArgs );

  #-------------------
  # Validate command..
  #-------------------
  if ( ! -x $sPsCmd ) {
    &logError ( 20126,
    "<%s> ps(1b) command [%s] is not executable", 
    $sMyName, $sPsCmd );
    return 0; # Failure..
  }

  #-------------------------------
  # Fire command and get details..
  #-------------------------------
  @hsProc { qw ( UID PID PPID COMMAND ) } =
  map {
    s/\n//;                             # Chomp..
    s/^\s+//;                           # Strip leading spaces..
    /^$sFmt$/o;                         # Extract fields..
  } readpipe ( "$sPsCmd $sPsCmdArgs" ); # From the command..

  $iExitCode = $? >> 8; # Get exit code..

  #------------------------
  # Restore env variables..
  #------------------------
  defined $sLang ? $ENV { LANG } = $sLang : delete $ENV { LANG };
  defined $sUnix95 ? $ENV { UNIX95 } = $sUnix95 : delete $ENV { UNIX95 };

  #------------------
  # Check exit code..
  #------------------
  if ( $iExitCode ) {
    &logError ( 20127,
    "<%s> Cmd [%s] with args [%s] failed with exit code [%s]",
    $sMyName, $sPsCmd, $sPsCmdArgs, $iExitCode );

    %hsProc = (); # Truncate..
    return 0;
  }

  #---------------------
  # Fill up the bucket..
  #---------------------
  push ( @$rlProcList, \%hsProc ); # Update..

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : bool GetSrvrDetails ( IN String       InitFile,
                                  OUT RefHashList SrvrDetails );

 Description : Gives PID file and process string for given server type

 Parameters  :

 $sInitFile                 The server initialization file for the Documentum connection broker
 $rhlSrvrDetails            Stores details of PID file and process string
 
 Returns : boolean value in the set of {0,1} where
           0 false
           1 true
------------------------------------------------------------------------------------------------------

=cut

sub GetSrvrDetails {
  #----------------
  # Get Arguments..
  #----------------
  my ( $sInitFile,
       $rhlSrvrDetails ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20128, 
  "Subroutine <%s> called with:
  InitFile                 [%s]
  SrvrDetails              [%s]",
  $sMyName, $sInitFile, $rhlSrvrDetails );

  #-----------------------------
  # Set return code to success..
  #-----------------------------
  my $iReturnCode = 1; # Set to true..

  #-------------------
  # Define variables..
  #-------------------
  my ( $sKey, $sProc, $sPidFile, $sPIDDir, $sResourceType,
       $sResourceName, $iPat ) = ();

  #-------------------------------------
  # List of search patterns to monitor..
  #-------------------------------------
  my @lsSrchPatterns = ( qr!dmdocbroker(.)*-init_file $sInitFile! );

  #----------------------------------------------
  # Derive our resource type and resource names..
  #----------------------------------------------
  $sResourceType = ( $0 =~ m@([^/]+)@g ) [-2];
  $sResourceName   = $ENV { VCS_LOG_RESOURCE_NAME };
  $sPIDDir = "/tmp/.VRTS$sResourceType/$sResourceName/PID";

  #------------------------------
  # Create PID folder if needed..
  #------------------------------
  if ( -d $sPIDDir ) {
    &logTrace ( 20129, "<%s> PID folder [%s] exists",
    $sMyName, $sPIDDir );
  }
  elsif ( ! mkpath ( $sPIDDir, 0, 0700 ) ) {
    &logWarn ( 20130, "<%s> Failed to create PID folder [%s]",
    $sMyName, $sPIDDir );
    return 0; # Cannot proceed..
  }

  %$rhlSrvrDetails = (); # Truncate..

  #-----------------------------------------------
  # Set appropriate filter strings and PID files..
  #-----------------------------------------------
  foreach $sProc ( @lsSrchPatterns ) {
    $sKey = 'srchstr_' . ++ $iPat;
    $sPidFile = "$sPIDDir/pidfile.$sKey";
    @{ $$rhlSrvrDetails { $sKey } } = ( $sPidFile, $sProc ); # Store..
  }

  return $iReturnCode;
}

=pod

------------------------------------------------------------------------------------------------------
 Function : int GetEngineTimeout ( IN Bool UseMargin );

 Description : Calculates engine timeout for entry point
 
 Parameters  :

 $bUseMargin      Flag to indicate if safety margin is requested

                  boolean value in the set of {0,1} where
                  0 No safety margin needed
                  1 Safety margin to be used
 Returns:  Time in secs.
------------------------------------------------------------------------------------------------------

=cut

sub GetEngineTimeout {
  #----------------
  # Get Arguments..
  #----------------
  my ( $bUseMargin ) = @_;

  #-----------------
  # Announce Entry..
  #-----------------
  my $sMyName = ( caller 0 )[3];

  &logTrace ( 20131, 
  "Subroutine <%s> called with: 
  UseMargin  [%s]",
  $sMyName, $bUseMargin );

  #-------------------------
  # Define local variables..
  #-------------------------
  my ( $sAttribute, $sCommand, $sCommandArgs,
       $iEngineTimeout, $iExitCode ) = ();

  #----------------------------------
  # Grace margins from VCS timeouts..
  #----------------------------------
  my %hiGraceMargin = ( 'MonitorTimeout', 10,   # Seconds..
                        'OnlineTimeout',  10,
                        'OfflineTimeout', 10,
                        'CleanTimeout',   20 );

  #--------------------------------------------------
  # Attempt to get resource level overridden values..
  #--------------------------------------------------
  my $sResourceName = $ENV { VCS_LOG_RESOURCE_NAME };
  my ( $sEntryPointName, $sResourceType ) = reverse $0 =~ m@([^/]+)@g;
  ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;

  #-----------------------------------
  # Query if it has already been set..
  #-----------------------------------
  if ( defined $::giEPTimeout &&    # Defined..
       $::giEPTimeout =~ /^\d+/ ) { # to an integer..
    $iEngineTimeout = $::giEPTimeout;

    &logTrace ( 20132,
    "<%s> Using AEPTimeout [%s] as EP timeout value",
    $sMyName, $iEngineTimeout );
  }
  else {
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    $sCommand = $sCmdHares; # Get global..
    $sCommandArgs = "-value $sResourceName $sAttribute";
  
    #----------------------------------
    # Fire command and collect output..
    #----------------------------------
    ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs 2>/dev/null" ) )
                      =~ s/\n//;
  
    #--------------------
    # Check exit status..
    #--------------------
    if ( $iExitCode = $? >> 8 ) {
      &logTrace ( 20133,
      "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
      $sMyName, $sCommand, $sCommandArgs, $iExitCode );
    }
    elsif ( ! $iEngineTimeout ) { # Undefined value..
      &logTrace ( 20134, 
      "<%s> Could not get [%s] from engine",
      $sMyName, $sAttribute );
    }
    else { # Success..
      &logTrace ( 20135, 
      "<%s> Got [%s] as [%s] sec from engine",
      $sMyName, $sAttribute, $iEngineTimeout );
    }
  
    #-----------------------------------------
    # Form query string based on entry point..
    #-----------------------------------------
    unless ( $iEngineTimeout ) {
      $sCommand = $sCmdHatype; # Get global..
      ( $sAttribute = $sEntryPointName ) =~ s/(.*)/\u$1Timeout/;
      $sResourceType = "ConnectionBroker" if $sResourceType !~ /\w/; # Ensure..
      $sCommandArgs = "-value $sResourceType $sAttribute";
    
      #----------------------------------
      # Fire command and collect output..
      #----------------------------------
      ( $iEngineTimeout = readpipe ( "$sCommand $sCommandArgs" ) ) =~ s/\n//;
    
      #--------------------
      # Check exit status..
      #--------------------
      if ( $iExitCode = $? >> 8 ) {
        &logTrace ( 20136,
        "<%s> Command [%s] with Args [%s] failed with exitcode [%s]",
        $sMyName, $sCommand, $sCommandArgs, $iExitCode );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      elsif ( ! $iEngineTimeout ) { # Empty value..
        &logTrace ( 20137, 
        "<%s> Could not get [%s] from engine",
        $sMyName, $sAttribute );
        $iEngineTimeout = $bUseMargin ? $hiGraceMargin { $sAttribute } : 0;
      }
      else { # Success..
        &logTrace ( 20138, 
        "<%s> Got [%s] as [%s] sec from engine",
        $sMyName, $sAttribute, $iEngineTimeout );
      }
    }
  }

  #----------------------------
  # Apply margin if requested.. 
  #----------------------------
  if ( $bUseMargin     &&                                    # Margin requested..
       $iEngineTimeout >= $hiGraceMargin { $sAttribute } ) { # And valid..
    $iEngineTimeout -= $hiGraceMargin { $sAttribute };       # Apply..

    &logTrace ( 20139,
    "<%s> Setting up [%s] sec margin. Returning [%s]",
    $sMyName, $hiGraceMargin { $sAttribute }, $iEngineTimeout );
  }

  return ( $iEngineTimeout ); # Secs..
}

1; # So PERL knows its included..
